unit ColorScheme;
// Project "Cole Research Libraries"
// Copyright  2002 Cole Research.  All Rights Reserved.
//
// SUBSYSTEM:     Cole Research Standard Libraries
// AUTHOR:        Jay Cole
// WRITTEN:       05/12/2002
// TEST STUB:     TestColorConfig.dpr <- Via DUNIT.
// STATUS:        Creating
// VERSION:       1.0 (walking thru)
//                                                                             
// OVERVIEW                                                                    
// ========                                                                    
// Allows a developer to easily add a color scheme configuration element to a 
// form.  This handles mutliple values for different display elements in a single
// component.  Essentially simulates the color setup of Delphi.
//
// NOTES
// =====
//
// SUGGESTED ENHANCEMENTS
// ======================
//
// UPDATE HISTORY
// ==============
// 05/10/02 (jlc) - Created initial object.
//

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   ExtCtrls, StdCtrls, ColorGrd;

type
   // Internally used data structures for holding the display elements attributes.
   // Inserted into a string list as objects.
   TDisplayElementInfo = class(TObject)
   private
      fBGColor : TColor;
      fFGColor : TColor;
      fStyle : TFontStyles;
   published
      property BGColor : TColor read fBGColor write fBGColor;
      property FGColor : TColor read fFGColor write fFGColor;
      property Style : TFontStyles read fStyle write fStyle;
   end;

   // Remember to put this in so they can access independent of the visual 
   // control later.
   TElementList = class(TStringList)
   private 
      fRegistryPath : string;
      fAssocListBox : TListBox;

   public
      constructor Create(aRegistryPath : string; anAssocListBox : TListBox); 
      destructor Destroy; override;
      procedure Save;
      procedure Reset;
      procedure Read;
      function DisplayInfo(ndx : integer) : TDisplayElementInfo;
      function DisplayInfoByName(name : string) : TDisplayElementInfo;
   end;

   
   // Event methods.
   TOnSelectElement = procedure(oldNdx, newNdx : integer) of object;
   TColorRange = 0..19;

   // Color scheme implementation.
   TColorScheme = class(TCustomPanel)
   private
      // Other fields
      fCopyright : string;
      fRegistryPath : string;
      fDefaultBGIndex : TColorRange;
      fDefaultFGIndex : TColorRange;
      fDefaultFontStyle : TFontStyles;

      // State variables
      fChanging : boolean;

      // Internal lookups.  fElements is redundant, but necessary because
      // child controls are wiped out prior to Destroy being called.
      fColorMap : array[0..ColorGrd.NumPaletteEntries - 1] of TPaletteEntry;
      fElements : TElementList;

      // Internal procedures
      procedure SetSubControls;
      procedure InternalResize(Sender: TObject);   
      procedure InternalCanResize(Sender: TObject; var NewWidth, NewHeight: Integer; var Resize: Boolean); 
      procedure ScreenToObject;
      procedure ObjectToScreen;
      procedure ControlChange;

      // Event handlers
      procedure HandleClick(Sender: TObject);
      procedure HandleKeyPress(Sender: TObject; var Key: Char);      
      procedure HandleChange(Sender: TObject);

      procedure SetRegistryPath(value : string);
            
   protected
      // Sub controls (list box, color select, font attr
      fDisplayElementsLB : TListBox;

      fColorPNL : TGroupBox;
      fColorSelectorCGRD : TColorGrid;

      fFontPNL : TGroupBox;
      fBoldCHK : TCheckBox;
      fItalicsCHK : TCheckBox;
      fUnderlineCHK : TCheckBox;

      fExampleLBL : TLabel;

   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      function AddElement(elementName : string) : integer;
      procedure DeleteElement(elementName : string);
      procedure DeleteElementAt(ndx : integer);
      procedure Save;
      
   published
      property Font;
      property Color;
      property Enabled;
      property ParentColor;
      property ParentShowHint;
      property Visible;
      property Left;
      property Top;
      property Height default 275;
      property Width default 350;
      
      property Copyright : string read fCopyright write fCopyright;
      property RegistryPath : string read fRegistryPath write SetRegistryPath;

      // These correspond to the 0..15 in the color grid instead of TColor stuff.
      property DefaultBGIndex : TColorRange read fDefaultBGIndex write fDefaultBGIndex default 0;
      property DefaultFGIndex : TColorRange read fDefaultFGIndex write fDefaultFGIndex default 15;
      property DefaultFontStyle : TFontStyles read fDefaultFontStyle write fDefaultFontStyle;
      property Elements : TElementList read fElements;
   end;

procedure Register;

implementation
uses Registry;



////////////////////////////////////////////////////////////////////////////////
//                       TColorScheme Public Methods                          //
////////////////////////////////////////////////////////////////////////////////

// (walkthru) Creates the initial state of the component.  Remember, this 
// component has several sub-components.
constructor TColorScheme.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);

   // Set the field names
   Copyright := 'Copyright  2002 by Cole Research';
   Caption := '';

   // Other properties, default behavior.
   BevelOuter := bvNone;
   BorderWidth := 0;
   BorderStyle := bsNone;
   Height := 275;
   Width := 350;
   DefaultBGIndex := 0;
   DefaultFGIndex := 15;
   DefaultFontStyle := [];

   // Create the sub objects.  Don't resize yet, we have another routine call
   // to do that because resizing at design time can create need too.
   // InsertControl is necessary to display to component, otherwise it never
   // appears.
   fDisplayElementsLB := TListBox.Create(self);
   InsertControl(fDisplayElementsLB);
   fDisplayElementsLB.OnKeyPress := HandleKeyPress;
   fDisplayElementsLB.OnClick := HandleClick;

   fColorPNL := TGroupBox.Create(self);
   fColorPNL.Caption := ' &Current Color ';
   InsertControl(fColorPNL);
   fColorSelectorCGRD := TColorGrid.Create(fColorPNL);
   fColorPNL.InsertControl(fColorSelectorCGRD);
   fColorSelectorCGRD.OnChange := HandleChange;

   fFontPNL := TGroupBox.Create(self);
   fFontPNL.Caption := ' &Text Attributes ';
   InsertControl(fFontPNL);
   fBoldCHK := TCheckBox.Create(fFontPNL);
   fFontPNL.InsertControl(fBoldCHK);
   fBoldCHK.Caption := '&Bold';
   fBoldCHK.OnClick := HandleChange;
   fItalicsCHK := TCheckBox.Create(fFontPNL);
   fFontPNL.InsertControl(fItalicsCHK);
   fItalicsCHK.Caption := '&Italics';
   fItalicsCHK.OnClick := HandleChange;
   fUnderlineCHK := TCheckBox.Create(fFontPNL);
   fFontPNL.InsertControl(fUnderlineCHK);
   fUnderlineCHK.Caption := '&Underline';
   fUnderlineCHK.OnClick := HandleChange;

   fExampleLBL := TLabel.Create(self);
   fExampleLBL.Caption := 'Example Text';
   fExampleLBL.AutoSize := false;
   InsertControl(fExampleLBL);
   SetSubControls;

   // Don't change palette's without checking through the entire source...this
   // is highly coupled with the TColorGrd components.
   GetPaletteEntries(GetStockObject(DEFAULT_PALETTE), 0, ColorGrd.NumPaletteEntries, fColorMap);
   fElements := TElementList.Create(fRegistryPath, fDisplayElementsLB);

   // Changing variable allows us to change the value of controls without 
   // appearing to be the user themselves, creating an infinite loop.
   fChanging := false;
   OnResize := InternalResize;
   OnCanResize := InternalCanResize;
end;


// (walkthru) Pretty simple, just get rid of the elements stuff.
destructor TColorScheme.Destroy;
begin
   // Free up string elements.  Don't use clear because it includes the now,
   // already dead listbox.
   fElements.Free;
   inherited Destroy;
end;


// (debugged) Adds a new element to the list per the default settings in the 
// master component.
function TColorScheme.AddElement(elementName : string) : integer;
var
   newObj : TDisplayElementInfo;
begin
   newObj := TDisplayElementInfo.Create;
   newObj.BGColor := DefaultBGIndex;
   newObj.FGColor := DefaultFGIndex;
   newObj.Style := DefaultFontStyle; 
   result := fDisplayElementsLB.Items.Add(elementName);
   fElements.AddObject(elementName, newObj);
end;


// (walkthru) Get the index in the listbox which corresponds to the same #
// in the fElement structure.
procedure TColorScheme.DeleteElement(elementName : string);
var ndx : integer;
begin
   ndx := fDisplayElementsLB.Items.IndexOf(elementName);
   Assert(fDisplayElementsLB.Items[ndx] = fElements[ndx], 'Problem w/ syncronizing string lists.');
   DeleteElementAt(ndx);
end;


// (walkthru) Delete both stringlist elements.  Free up the object associated
// with fElement entry.
procedure TColorScheme.DeleteElementAt(ndx : integer);
begin
   fDisplayElementsLB.Items.Delete(ndx);
   fElements.DisplayInfo(ndx).Free;
   fElements.Delete(ndx);
end;


// (walkthru) A difficult choice arises here.  Keep the existing and then issue
// a clear command, or when the registry changes, assume they want to read from
// the registry and start with a clean slate.  We picked the latter.
procedure TColorScheme.SetRegistryPath(value : string);
begin
   if (value <> fRegistryPath) then begin
      fRegistryPath := value;
      fElements.fRegistryPath := value;

      // The clear and read also set our listbox string list at the same time...
      // so nothing here.
      fElements.Reset;
      fElements.Read;
   end;
end;


////////////////////////////////////////////////////////////////////////////////
//                       TColorScheme Private Methods                         //
////////////////////////////////////////////////////////////////////////////////

const
   cMinControlSize : integer = 120;

// (debugged) Save the elements to the registry.
procedure TColorScheme.Save;
begin
   fElements.Save;
end;


// (walkthru) Set the initial juxta-position of the controls w/ each other.
procedure TColorScheme.SetSubControls;
const
   cTopMargin = 2;
   cLBTopMargin = cTopMargin + 6;
   cTotMarginH = cLBTopMargin + 4;
   
   cLeftMargin = 2;
   cTotMarginW = cLeftMargin * 2;
   
   cGutterWLen = 8;
   cGutterHLen = 8;

   cPanelCompTop = 16;
   cPanelCompLeft = 10;
   
begin
   // 60% of the left hand side
   fDisplayElementsLB.Top := cLBTopMargin;
   fDisplayElementsLB.Left := cLeftMargin;
   fDisplayElementsLB.Width := (Width - cGutterWLen - cTotMarginW) * 6 div 10;
   fDisplayElementsLB.Height := (Height - cTotMarginH);

   // 40% of the right hand side
   fColorPNL.Top := cTopMargin;
   fColorPNL.Left := fDisplayElementsLB.Left + fDisplayElementsLB.Width +  cGutterWLen;
   fColorPNL.Width := (Width - cGutterWLen - cTotMarginW) * 4 div 10;
   fColorPNL.Height := (Height - cTotMarginH) div 2;

   // Now the internal controls.  Constants are margins for control.
   with fColorSelectorCGRD do begin
      Top := cPanelCompTop;
      Left := cPanelCompLeft;
      Width := fColorPNL.Width - Left * 2;
      Height := fColorPNL.Height - Top - 6; // <- Top higher for title = 6
   end;   

   // Allocate room for 3 checkboxes.
   fFontPNL.Top := fColorPNL.Top + fColorPNL.Height + cGutterHLen;
   fFontPNL.Left := fColorPNL.Left;
   fFontPNL.Width := fColorPNL.Width;
   fFontPNL.Height := (fBoldCHK.Height + 2) * 3 + 18;  // Allow for the borders in the PNL

   // Stack the checkboxes down the panel.
   fBoldCHK.Top := cPanelCompTop;
   fBoldCHK.Left := cPanelCompLeft;
   fItalicsCHK.Top := fBoldCHK.Top + fBoldCHK.Height + 2;
   fItalicsCHK.Left := cPanelCompLeft;
   fUnderlineCHK.Top := fItalicsCHK.Top + fItalicsCHK.Height + 2;
   fUnderlineCHK.Left := cPanelCompLeft;

   // Set the example and allow one more in Height for the underline.
   fExampleLBL.Width := fColorPNL.Width;
   fExampleLBL.Height := fExampleLBL.Height + 1;
   fExampleLBL.Left := fDisplayElementsLB.Left + fDisplayElementsLB.Width + cGutterWLen;
   fExampleLBL.Top := 
      fDisplayElementsLB.Top + fDisplayElementsLB.Height - fExampleLBL.Height;
end;


// (debugged) Just reset the subcontrols that proc calculates the new placements
// based on the size of the parent.
procedure TColorScheme.InternalResize(Sender: TObject);   
begin
   // Resize the child controls
   SetSubControls;
end;


// (walkthru) Just don't allow it to go below the minimum size.  This is needed
// to prevent the control from going too small and making the object unreadable.
procedure TColorScheme.InternalCanResize(Sender: TObject; var NewWidth, NewHeight: Integer; var Resize: Boolean); 
begin
   if (NewWidth < cMinControlSize) then NewWidth := cMinControlSize;
   if (NewHeight < cMinControlSize) then NewHeight := cMinControlSize;
   Resize := true;
end;


// (walkthru) Take the current highlighted item in the list box and get the
// associated object and then set up the screen controls appropriately.
procedure TColorScheme.ObjectToScreen;
var 
   anElem : TDisplayElementInfo;
   tmpInt : integer;
begin
   if (fDisplayElementsLB.ItemIndex >= 0) then begin
      anElem := fElements.DisplayInfo(fDisplayElementsLB.ItemIndex);
      with fColorSelectorCGRD do begin
         // We need to check for non-existant palette entries to reset them
         // if there is a problem.
         tmpInt := fColorSelectorCGRD.ColorToIndex(anElem.FGColor);
         if (tmpInt = -1) then ForegroundIndex := 0 else ForegroundIndex := tmpInt;
         tmpInt := fColorSelectorCGRD.ColorToIndex(anElem.BGColor);
         if (tmpInt = -1) then BackgroundIndex := 0 else BackgroundIndex := tmpInt;
         fBoldCHK.Checked := fsBold in anElem.Style;
         fItalicsCHK.Checked := fsItalic in anElem.Style;
         fUnderlineCHK.Checked := fsUnderline in anElem.Style;
      end;
   end;
end;


// (walkthru)  Takes what's on the screen and puts it into the current object
// structure associated with the list box item.
procedure TColorScheme.ScreenToObject;
var anElem : TDisplayElementInfo;
begin
   if (fDisplayElementsLB.ItemIndex >= 0) then begin
      anElem := fElements.DisplayInfo(fDisplayElementsLB.ItemIndex);
      anElem.BGColor := fExampleLBL.Color;
      anElem.FGColor := fExampleLBL.Font.Color;
      anElem.Style := fExampleLBL.Font.Style;
   end;
end;


// (walkthru) The edit controls change either through a user interaction w/ 
// them, or through a new element being selected.  This resets the example label
// to reflect the new color and style combination.
procedure TColorScheme.ControlChange;
var ndx : integer;
begin
   // Since color grd works on an index system and we work on RGB TColor, we
   // have to do this slight of hand via the Palette.  Color Grid also does 
   // something entirely stupid...it offset's by 4 if the palette is greater 
   // than 7.  !!! Change ColorGrd later to be smarter.
   ndx := fColorSelectorCGRD.BackgroundIndex;
   if (ndx > 7) then ndx := ndx + 4; 
   with fColorMap[ndx] do 
      fExampleLBL.Color := TColor(RGB(peRed, peGreen, peBlue));

   ndx := fColorSelectorCGRD.ForegroundIndex;
   if (ndx > 7) then ndx := ndx + 4; 
   with fColorMap[ndx] do 
      fExampleLBL.Font.Color := TColor(RGB(peRed, peGreen, peBlue));

   fExampleLBL.Font.Style := [];
   if (fBoldCHK.Checked) then fExampleLBL.Font.Style := fExampleLBL.Font.Style + [fsBold];
   if (fItalicsCHK.Checked) then fExampleLBL.Font.Style := fExampleLBL.Font.Style + [fsItalic];
   if (fUnderlineCHK.Checked) then fExampleLBL.Font.Style := fExampleLBL.Font.Style + [fsUnderline];
end;
   

// (debugged) Performs the necessary functions when the user clicks on a new
// list box entry.
procedure TColorScheme.HandleClick(Sender: TObject);
begin
   // fChanging prevents recursive changes.
   if (not fChanging) then begin
      fChanging := true;
      ObjectToScreen; 
      ControlChange;
      fChanging := false;  
   end;
end;


// (walkthru) If any of the controls changes value, this routine is called.
procedure TColorScheme.HandleChange(Sender: TObject);
begin
   // fChanging prevents recursive changes.
   if (not fChanging) then begin
      fChanging := true;
      ControlChange;
      ScreenToObject;
      fChanging := false;
   end;
end;


// (walkthru) If a key is pressed instead of a click on the colorgrid, we
// need to update the example. 
procedure TColorScheme.HandleKeyPress(Sender: TObject; var Key: Char);      
begin
   if (not fChanging) then begin
      fChanging := true;
      ObjectToScreen;
      fChanging := false;
   end;
end;


////////////////////////////////////////////////////////////////////////////////
//                         TElementList Implementation                        //
////////////////////////////////////////////////////////////////////////////////

const 
   // Registry key strings
   cElemCntStr : string = 'CSNumElems';
   cElemNameStr : string = 'CSElemStr';
   cElemFGStr : string = 'CSElemFG';
   cElemBGStr : string = 'CSElemBG';
   cElemStyleStr : string = 'CSElemStyleStr';


// (debugged) Pass in the registry path to save to and the associated visual 
// control where the element strings will be displayed.  The associated listbox
// can be nil for accessing elements and attributes outside of the control.
constructor TElementList.Create(aRegistryPath : string; anAssocListBox : TListBox); 
begin
   inherited Create;
   fRegistryPath := aRegistryPath;
   fAssocListBox := anAssocListBox;
   Read;
end;


// (walkthru) Free up the strings and objects.  Can't just call Reset because
// the listbox may not already be destroyed.
destructor TElementList.Destroy;
var ndx : integer;
begin
   for ndx := Count - 1 downto 0 do begin
      Objects[ndx].Free;
      Delete(ndx);
   end;
   inherited Destroy;
end;



// (walkthru) Read the strings and element attributes from the registry.
procedure TElementList.Read;
var 
   registry : TRegistry;
   ndx, numElems : integer;
   anElem : TDisplayElementInfo;
   elemName : string;
   elemFG, elemBG : TColor;
   elemStyle : TFontStyles;
begin
   // Read the results in
   if (fRegistryPath <> '') then begin
      // We have a registry entry.
      registry := nil;
      try
         registry := TRegistry.Create;

         // Remember, don't create the registry entry if none has been created
         // so far.  We don't want to create it explicitely because the user
         // has to add the keys initially.
         if registry.OpenKey(fRegistryPath, false) then begin
            numElems := registry.ReadInteger(cElemCntStr);
            for ndx := 0 to numElems - 1 do begin
               elemName := registry.ReadString(cElemNameStr+IntToStr(ndx));
               registry.ReadBinaryData(cElemFGStr+IntToStr(ndx), elemFG, SizeOf(elemFG));
               registry.ReadBinaryData(cElemBGStr+IntToStr(ndx), elemBG, SizeOf(elemBG));
               registry.ReadBinaryData(cElemStyleStr+IntToStr(ndx), elemStyle, SizeOf(elemStyle));

               anElem := TDisplayElementInfo.Create;
               AddObject(elemName, anElem);
               if (fAssocListBox <> nil) then fAssocListBox.Items.Add(elemName);
               anElem.FGColor := elemFG;
               anElem.BGColor := elemBG;
               anElem.Style := elemStyle;
            end;
         end;
      finally
         registry.CloseKey;
         registry.Free;
      end;      
   end;
end;


// (walkthru) Reset the list and also clear the strings in the associated list
// box.
procedure TElementList.Reset;
var ndx : integer;
begin
   for ndx := Count - 1 downto 0 do begin
      Objects[ndx].Free;
      Delete(ndx);
      if (fAssocListBox <> nil) then fAssocListBox.Items.Delete(ndx);
   end;
end;


// (walkthru) We don't want the destroy method to always update the registry, 
// what if the user presses cancel.  Write() allows us flush when we want.  Now,
// the system doesn't write to the registry upon free'ing.
procedure TElementList.Save;
var 
   registry : TRegistry;
   ndx : integer;
   anElem : TDisplayElementInfo;
begin
   // Write the results out.
   if (fRegistryPath <> '') then begin
      // We have a registry entry.
      registry := nil;
      try
         registry := TRegistry.Create;
         registry.OpenKey(fRegistryPath, true);

         registry.WriteInteger(cElemCntStr, Count);
         for ndx := 0 to Count - 1 do begin
            registry.WriteString(cElemNameStr+IntToStr(ndx), self[ndx]);
            anElem := TDisplayElementInfo(Objects[ndx]);
            registry.WriteBinaryData(cElemFGStr+IntToStr(ndx), anElem.fFGColor, SizeOf(anElem.FGColor));
            registry.WriteBinaryData(cElemBGStr+IntToStr(ndx), anElem.fBGColor, SizeOf(anElem.BGColor));
            registry.WriteBinaryData(cElemStyleStr+IntToStr(ndx), anElem.fStyle, SizeOf(anElem.Style));
         end;
      finally
         registry.CloseKey;
         registry.Free;
      end;      
   end;
end;


// (debugged) Just the object array encapsulated to return DisplayElementInfo
// data types so no need for a typecast.
function TElementList.DisplayInfo(ndx : integer) : TDisplayElementInfo;
begin
   result := TDisplayElementInfo(Objects[ndx]);      
end;


// (walkthru) Returns the elements associated with a name.  This is not case
// sensitive for no particularly good reason.  Returns a nil value for the 
// display element
function TElementList.DisplayInfoByName(name : string) : TDisplayElementInfo;
var ndx : integer;
begin
   ndx := IndexOf(name);
   if (ndx = -1) then
      result := nil
   else
      result := TDisplayElementInfo(Objects[ndx]);
end;



////////////////////////////////////////////////////////////////////////////////
//                            Registration Procedure                          //
////////////////////////////////////////////////////////////////////////////////

procedure Register;
begin
  RegisterComponents('Cole Research', [TColorScheme]);
end;

end.
